﻿using AiYoloV5Onnx.Common;
using Microsoft.ML.OnnxRuntime;
using Microsoft.ML.OnnxRuntime.Tensors;
using SkiaSharp;

namespace AiYoloV5Onnx.Yolo
{
    public class Yolov5Detector : IDisposable
    {
        private InferenceSession _OnnxSession { get; }

        private float _YoloWidth { get; }

        private float _YoloHeight { get; }

        private string[] _YoloLabes { get; }

        public Yolov5Detector(string modelPath, string tagPath, float width = 640, float height = 640, bool useDirectML = false)
        {
            var options = new Microsoft.ML.OnnxRuntime.SessionOptions();
            //options.LogSeverityLevel = OrtLoggingLevel.ORT_LOGGING_LEVEL_INFO;
            if (useDirectML)
            {
                options.AppendExecutionProvider_DML(0);
            }
            else
            {
                options.IntraOpNumThreads = 2;
                options.ExecutionMode = ExecutionMode.ORT_PARALLEL;
                options.InterOpNumThreads = 6;
                options.GraphOptimizationLevel = GraphOptimizationLevel.ORT_ENABLE_ALL;
                //options.OptimizedModelFilePath = model_path;
                options.AppendExecutionProvider_CPU(0);
            }
            // create inference session
            this._OnnxSession = new InferenceSession(modelPath, options);
            this._YoloWidth = width;
            this._YoloHeight = height;
            //
            var tagText = File.ReadAllText(tagPath);
            _YoloLabes = tagText.Split("\r\n");
        }

        private List<Prediction> GetInfo(SKBitmap img)
        {
            var width = Convert.ToInt32(_YoloWidth);
            var height = Convert.ToInt32(_YoloHeight);
            //
            //var resized_image = img.ResizeWithoutPadding(width, height);
            var resized_image = img.Resize(width, height);
            //计算张量
            var input_tensor = FastToOnnxTensor(resized_image);
            //
            var container = new List<NamedOnnxValue>();
            container.Add(NamedOnnxValue.CreateFromTensor("images", input_tensor));
            //
            using var results = _OnnxSession.Run(container);
            var resultsArray = results.ToArray();
            var tensors = resultsArray[0].AsTensor<float>();
            var array = tensors.ToArray();
            var preds = ParseResults(array);
            //
            return preds;
        }

        public List<Prediction> Inference(SKBitmap bitmap)
        {
            var width = Convert.ToInt32(_YoloWidth);
            var height = Convert.ToInt32(_YoloHeight);
            //
            var preds = GetInfo(bitmap);
            //绘制
            bitmap.DrawYoloPrediction(preds, SKColors.Red, SKColors.Red, width, height);
            //图片
            return preds;
        }

        public Task<List<Prediction>> InferenceAsync(SKBitmap bitmap)
        {
            return Task.Run(() =>
            {
                lock (_OnnxSession)
                {
                    return Inference(bitmap);
                }
            });
        }

        /// <summary>
        /// 计算张量（RRRR GGGG BBBB）
        /// </summary>
        /// <param name="source">图像</param>
        /// <returns></returns>
        public static Tensor<float> FastToOnnxTensor(SKBitmap source)
        {
            var tensorShape = new int[] { 1, 3, source.Height, source.Width };
            var tensor = new DenseTensor<float>(tensorShape);

            for (int y = 0; y < source.Height; y++)
            {
                for (int x = 0; x < source.Width; x++)
                {
                    var pixel = source.GetPixel(x, y);
                    tensor[0, 0, y, x] = pixel.Red / 255f; // R
                    tensor[0, 1, y, x] = pixel.Green / 255f; // G
                    tensor[0, 2, y, x] = pixel.Blue / 255f; // B
                }
            }

            return tensor;
            //var totalPixelsCount = source.Width * source.Height;
            //var floatArray = new float[source.Width * source.Height * 3];
            //var imageInfo = source.Info;
            ////将 IntPtr 转换为字节数组
            //var bytes = new byte[imageInfo.BytesSize];
            //Marshal.Copy(source.GetPixels(), bytes, 0, bytes.Length);
            ////多线程并行循环
            //Parallel.For(0, totalPixelsCount, (pIdx, state) =>
            //{
            //    //RGB 起始索引
            //    var r_idx = pIdx + totalPixelsCount * 0;
            //    var g_idx = pIdx + totalPixelsCount * 1;
            //    var b_idx = pIdx + totalPixelsCount * 2;
            //    //每个像素由4个字节组成（ARGB）
            //    var pixelOffset = pIdx * 4;
            //    //SkiaSharp 使用预乘alpha，所以我们需要除以alpha通道值（A）来得到正确的RGB值
            //    var alpha = bytes[pixelOffset + 3] / 255f;
            //    floatArray[r_idx] = bytes[pixelOffset + 2] / 255f * alpha; // R
            //    floatArray[g_idx] = bytes[pixelOffset + 1] / 255f * alpha; // G
            //    floatArray[b_idx] = bytes[pixelOffset + 0] / 255f * alpha; // B
            //});
            ////
            //var memory = new Memory<float>(floatArray);
            //var dimensions = new int[] { 1, 3, source.Height, source.Width };
            //var denseTensor = new DenseTensor<float>(memory, dimensions);
            //return denseTensor;
        }

        /// <summary>
        /// 解析结果
        /// </summary>
        /// <param name="results">计算后的张量</param>
        /// <returns></returns>
        private List<Prediction> ParseResults(float[] results)
        {
            //边界框长度
            int confidenceIndex = 4;
            //置信度长度
            int labelStartIndex = 5;
            //标签数量
            int labelLength = _YoloLabes.Length;
            //每个分区长度
            int dimensions = labelLength + 5;
            //分区数量
            int rows = results.Length / dimensions;
            //
            var detections = new List<Prediction>();
            for (int i = 0; i < rows; ++i)
            {
                var index = i * dimensions;
                //不要置信度低的
                if (results[index + confidenceIndex] <= 0.4f) continue;
                //对每个预测类别的置信度进行缩放
                for (int j = labelStartIndex; j < dimensions; ++j)
                {
                    results[index + j] = results[index + j] * results[index + confidenceIndex];
                }
                //分析每个预测类别
                for (int k = labelStartIndex; k < dimensions; ++k)
                {
                    //不要过低的预测类别
                    if (results[index + k] <= 0.5f) continue;

                    var value_0 = results[index];
                    var value_1 = results[index + 1];
                    var value_2 = results[index + 2];
                    var value_3 = results[index + 3];

                    var bbox = new BBox(
                        (value_0 - value_2 / 2) / _YoloWidth,
                        (value_1 - value_3 / 2) / _YoloWidth,
                        (value_0 + value_2 / 2) / _YoloHeight,
                        (value_1 + value_3 / 2) / _YoloHeight);

                    var l_index = k - labelStartIndex;
                    detections.Add(new Prediction()
                    {
                        Box = bbox,
                        Confidence = results[index + k],
                        LabelIndex = l_index,
                        LabelName = _YoloLabes[l_index]
                    });
                }
            }

            return Prediction.NMS(detections);
        }

        public void Dispose()
        {
            _OnnxSession?.Dispose();
        }
    }
}